# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Event handling logic for more info."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("EventManagerImpl",)

import asyncio
import base64
import logging
import random
import typing

from hikari import errors
from hikari import intents as intents_
from hikari import presences as presences_
from hikari import snowflakes
from hikari.api import config
from hikari.events import application_events
from hikari.events import auto_mod_events
from hikari.events import channel_events
from hikari.events import guild_events
from hikari.events import interaction_events
from hikari.events import member_events
from hikari.events import message_events
from hikari.events import monetization_events
from hikari.events import poll_events
from hikari.events import reaction_events
from hikari.events import role_events
from hikari.events import scheduled_events
from hikari.events import shard_events
from hikari.events import stage_events
from hikari.events import typing_events
from hikari.events import user_events
from hikari.events import voice_events
from hikari.impl import event_manager_base
from hikari.internal import time
from hikari.internal import ux

if typing.TYPE_CHECKING:
    from hikari import guilds
    from hikari import invites
    from hikari import voices
    from hikari.api import cache as cache_
    from hikari.api import entity_factory as entity_factory_
    from hikari.api import event_factory as event_factory_
    from hikari.api import shard as gateway_shard
    from hikari.internal import data_binding


_LOGGER: typing.Final[logging.Logger] = logging.getLogger("hikari.event_manager")


def _fixed_size_nonce() -> str:
    # This generates nonces of length 28 for use in member chunking.
    head = time.time_ns().to_bytes(8, "big")
    tail = random.getrandbits(92).to_bytes(12, "big")
    return base64.b64encode(head + tail).decode("ascii")


async def _request_guild_members(
    shard: gateway_shard.GatewayShard,
    guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
    *,
    include_presences: bool,
    nonce: str,
) -> None:
    try:
        await shard.request_guild_members(guild, include_presences=include_presences, nonce=nonce)

    # Ignore errors raised by a shard shutting down
    except errors.ComponentStateConflictError:
        pass


class EventManagerImpl(event_manager_base.EventManagerBase):
    """Provides event handling logic for Discord events."""

    __slots__: typing.Sequence[str] = (
        "_auto_chunk_members",
        "_cache",
        "_entity_factory",
        "_guild_member_request_tasks",
    )

    def __init__(
        self,
        entity_factory: entity_factory_.EntityFactory,
        event_factory: event_factory_.EventFactory,
        intents: intents_.Intents,
        /,
        *,
        auto_chunk_members: bool = True,
        cache: cache_.MutableCache | None = None,
    ) -> None:
        self._cache = cache
        self._auto_chunk_members = auto_chunk_members
        self._entity_factory = entity_factory
        self._guild_member_request_tasks: set[asyncio.Task[None]] = set()

        components = cache.settings.components if cache else config.CacheComponents.NONE
        super().__init__(event_factory=event_factory, intents=intents, cache_components=components)

    def _cache_enabled_for(self, components: config.CacheComponents, /) -> bool:
        return self._cache is not None and (self._cache.settings.components & components) == components

    @event_manager_base.filtered(shard_events.ShardReadyEvent, config.CacheComponents.ME)
    def on_ready(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#ready for more info."""
        # TODO: cache unavailable guilds on startup, I didn't bother for the time being.
        event = self._event_factory.deserialize_ready_event(shard, payload)

        if self._cache:
            self._cache.update_me(event.my_user)

        self.dispatch(event)

    @event_manager_base.filtered(shard_events.ShardResumedEvent)
    def on_resumed(self, shard: gateway_shard.GatewayShard, _: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#resumed for more info."""
        self.dispatch(self._event_factory.deserialize_resumed_event(shard))

    @event_manager_base.filtered(application_events.ApplicationCommandPermissionsUpdateEvent)
    def on_application_command_permissions_update(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        self.dispatch(self._event_factory.deserialize_application_command_permission_update_event(shard, payload))

    @event_manager_base.filtered(channel_events.GuildChannelCreateEvent, config.CacheComponents.GUILD_CHANNELS)
    def on_channel_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#channel-create for more info."""
        event = self._event_factory.deserialize_guild_channel_create_event(shard, payload)

        if self._cache:
            self._cache.set_guild_channel(event.channel)

        self.dispatch(event)

    @event_manager_base.filtered(channel_events.GuildChannelUpdateEvent, config.CacheComponents.GUILD_CHANNELS)
    def on_channel_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#channel-update for more info."""
        old = self._cache.get_guild_channel(snowflakes.Snowflake(payload["id"])) if self._cache else None
        event = self._event_factory.deserialize_guild_channel_update_event(shard, payload, old_channel=old)

        if self._cache:
            self._cache.update_guild_channel(event.channel)

        self.dispatch(event)

    @event_manager_base.filtered(
        channel_events.GuildChannelDeleteEvent,
        config.CacheComponents.GUILD_CHANNELS | config.CacheComponents.GUILD_THREADS,
    )
    def on_channel_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#channel-delete for more info."""
        event = self._event_factory.deserialize_guild_channel_delete_event(shard, payload)

        if self._cache:
            self._cache.delete_guild_channel(event.channel.id)
            self._cache.clear_threads_for_channel(event.guild_id, event.channel.id)

        self.dispatch(event)

    @event_manager_base.filtered((channel_events.GuildPinsUpdateEvent, channel_events.DMPinsUpdateEvent))
    def on_channel_pins_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#channel-pins-update for more info."""
        # TODO: we need a method for this specifically
        self.dispatch(self._event_factory.deserialize_channel_pins_update_event(shard, payload))

    @event_manager_base.filtered(
        (channel_events.GuildThreadAccessEvent, channel_events.GuildThreadCreateEvent),
        config.CacheComponents.GUILD_THREADS,
    )
    def on_thread_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#thread-create for more info."""
        event: channel_events.GuildThreadAccessEvent | channel_events.GuildThreadCreateEvent
        if "newly_created" in payload:
            event = self._event_factory.deserialize_guild_thread_create_event(shard, payload)

        else:
            event = self._event_factory.deserialize_guild_thread_access_event(shard, payload)

        if self._cache:
            self._cache.set_thread(event.thread)

        self.dispatch(event)

    @event_manager_base.filtered(channel_events.GuildThreadUpdateEvent, config.CacheComponents.GUILD_THREADS)
    def on_thread_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#thread-update for more info."""
        old = self._cache.get_thread(snowflakes.Snowflake(payload["id"])) if self._cache else None
        event = self._event_factory.deserialize_guild_thread_update_event(shard, payload, old_thread=old)

        if self._cache:
            self._cache.update_thread(event.thread)

        self.dispatch(event)

    @event_manager_base.filtered(channel_events.GuildThreadDeleteEvent, config.CacheComponents.GUILD_THREADS)
    def on_thread_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#thread-delete for more info."""
        event = self._event_factory.deserialize_guild_thread_delete_event(shard, payload)

        if self._cache:
            self._cache.delete_thread(event.thread_id)

        self.dispatch(event)

    @event_manager_base.filtered(channel_events.ThreadListSyncEvent, config.CacheComponents.GUILD_THREADS)
    def on_thread_list_sync(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#thread-list-sync for more info."""
        event = self._event_factory.deserialize_thread_list_sync_event(shard, payload)

        if self._cache:
            if event.channel_ids:
                for channel_id in event.channel_ids:
                    self._cache.clear_threads_for_channel(event.guild_id, channel_id)

            else:
                self._cache.clear_threads_for_guild(event.guild_id)

            for thread in event.threads.values():
                self._cache.set_thread(thread)

        self.dispatch(event)

    @event_manager_base.filtered(channel_events.ThreadMembersUpdateEvent, config.CacheComponents.GUILD_THREADS)
    def on_thread_members_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#thread-members-update for more info."""
        event = self._event_factory.deserialize_thread_members_update_event(shard, payload)

        if self._cache:
            user_id = event.shard.get_user_id()

            # We only care about us being removed here. When we are added, we will receive a THREAD_CREATE with
            # all the other info.
            if user_id in event.removed_member_ids:
                self._cache.delete_thread(event.thread_id)

        self.dispatch(event)

    # Internal granularity is preferred for GUILD_CREATE over decorator based filtering due to its large scope.
    def on_guild_create(  # noqa: C901, PLR0912, PLR0915 - Function too complex, too many branches and too long
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-create for more info."""
        event: guild_events.GuildAvailableEvent | guild_events.GuildJoinEvent | None

        unavailable = payload.get("unavailable")

        if unavailable:
            # It is possible for GUILD_CREATE to contain unavailable guilds during outages
            # In cases like this, we can just ignore the event and wait for the outage to
            # be resolved, where the correct guild visibility event will be dispatched
            return

        if unavailable is not None and self._enabled_for_event(guild_events.GuildAvailableEvent):
            event = self._event_factory.deserialize_guild_available_event(shard, payload)
        elif unavailable is None and self._enabled_for_event(guild_events.GuildJoinEvent):
            event = self._event_factory.deserialize_guild_join_event(shard, payload)
        else:
            event = None

        if event:
            # We also filter here to prevent iterating over them and calling a function that won't do anything
            channels = event.channels if self._cache_enabled_for(config.CacheComponents.GUILD_CHANNELS) else None
            emojis = event.emojis if self._cache_enabled_for(config.CacheComponents.EMOJIS) else None
            stickers = event.stickers if self._cache_enabled_for(config.CacheComponents.GUILD_STICKERS) else None
            guild = event.guild if self._cache_enabled_for(config.CacheComponents.GUILDS) else None
            guild_id = event.guild.id
            members = event.members if self._cache_enabled_for(config.CacheComponents.MEMBERS) else None
            presences = event.presences if self._cache_enabled_for(config.CacheComponents.PRESENCES) else None
            roles = event.roles if self._cache_enabled_for(config.CacheComponents.ROLES) else None
            voice_states = event.voice_states if self._cache_enabled_for(config.CacheComponents.VOICE_STATES) else None
            threads = event.threads if self._cache_enabled_for(config.CacheComponents.GUILD_THREADS) else None

        elif self._cache:
            _LOGGER.log(ux.TRACE, "Skipping on_guild_create dispatch due to lack of any registered listeners")
            gd = self._entity_factory.deserialize_gateway_guild(payload, user_id=shard.get_user_id())

            channels = gd.channels() if self._cache_enabled_for(config.CacheComponents.GUILD_CHANNELS) else None
            emojis = gd.emojis() if self._cache_enabled_for(config.CacheComponents.EMOJIS) else None
            stickers = gd.stickers() if self._cache_enabled_for(config.CacheComponents.GUILD_STICKERS) else None
            guild = gd.guild() if self._cache_enabled_for(config.CacheComponents.GUILDS) else None
            guild_id = gd.id
            members = gd.members() if self._cache_enabled_for(config.CacheComponents.MEMBERS) else None
            presences = gd.presences() if self._cache_enabled_for(config.CacheComponents.PRESENCES) else None
            roles = gd.roles() if self._cache_enabled_for(config.CacheComponents.ROLES) else None
            voice_states = gd.voice_states() if self._cache_enabled_for(config.CacheComponents.VOICE_STATES) else None
            threads = gd.threads() if self._cache_enabled_for(config.CacheComponents.GUILD_THREADS) else None

        else:
            _LOGGER.log(
                ux.TRACE, "Skipping on_guild_create raw dispatch due to lack of any registered listeners or cache need"
            )

            channels = None
            emojis = None
            stickers = None
            guild = None
            guild_id = snowflakes.Snowflake(payload["id"])
            members = None
            presences = None
            roles = None
            voice_states = None
            threads = None

        if self._cache:
            if guild:
                self._cache.update_guild(guild)

            if channels:
                self._cache.clear_guild_channels_for_guild(guild_id)
                for channel in channels.values():
                    self._cache.set_guild_channel(channel)

            if emojis:
                self._cache.clear_emojis_for_guild(guild_id)
                for emoji in emojis.values():
                    self._cache.set_emoji(emoji)

            if stickers:
                self._cache.clear_stickers_for_guild(guild_id)
                for sticker in stickers.values():
                    self._cache.set_sticker(sticker)

            if roles:
                self._cache.clear_roles_for_guild(guild_id)
                for role in roles.values():
                    self._cache.set_role(role)

            if members:
                # TODO: do we really want to invalidate these all after an outage.
                self._cache.clear_members_for_guild(guild_id)
                if not self._cache.settings.only_my_member:
                    for member in members.values():
                        self._cache.set_member(member)
                else:
                    my_member = members[shard.get_user_id()]
                    self._cache.set_member(my_member)

            if presences:
                self._cache.clear_presences_for_guild(guild_id)
                for presence in presences.values():
                    self._cache.set_presence(presence)

            if voice_states:
                self._cache.clear_voice_states_for_guild(guild_id)
                for voice_state in voice_states.values():
                    self._cache.set_voice_state(voice_state)

            if threads:
                self._cache.clear_threads_for_guild(guild_id)
                for thread in threads.values():
                    self._cache.set_thread(thread)

        # We only want to chunk if we are allowed and need to:
        #   Allowed?
        #       All the following must be true:
        #           1. [`auto_chunk_members`][] is true (the user wants us to).
        #           2. We have the necessary intents ([`GUILD_MEMBERS`][]).
        #           3. The guild is marked as "large" or we do not have [`GUILD_PRESENCES`][] intent
        #              Discord will only send every other member objects on the `GUILD_CREATE`
        #              payload if presence intents are also declared, so if this isn't the case then we also
        #              want to chunk small guilds.
        #
        #   Need to?
        #       One of the following must be true:
        #           1. We have a cache, and it requires it (it is enabled for [`MEMBERS`][]), but we are
        #              not limited to only our own member (which is included in the `GUILD_CREATE` payload).
        #           2. The user is waiting for the member chunks (there is an event listener for it).
        presences_declared = bool(self._intents & intents_.Intents.GUILD_PRESENCES)

        if (
            self._auto_chunk_members
            and self._intents & intents_.Intents.GUILD_MEMBERS
            and (payload.get("large") or not presences_declared)
            and (
                (
                    self._cache
                    and self._cache_enabled_for(config.CacheComponents.MEMBERS)
                    and not self._cache.settings.only_my_member
                )
                # This call is a bit expensive, so best to do it last
                or self._enabled_for_event(shard_events.MemberChunkEvent)
            )
        ):
            nonce = f"{shard.id}.{_fixed_size_nonce()}"

            if event:
                event.chunk_nonce = nonce

            # We create a task here instead of awaiting the result to avoid any rate-limits from delaying dispatch.
            task = asyncio.create_task(
                _request_guild_members(shard, guild_id, include_presences=presences_declared, nonce=nonce),
                name=f"{shard.id}:{guild_id} guild create members request",
            )

            self._guild_member_request_tasks.add(task)
            task.add_done_callback(self._guild_member_request_tasks.discard)

        if event:
            self.dispatch(event)

    # Internal granularity is preferred for GUILD_UPDATE over decorator based filtering due to its large scope.
    def on_guild_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-update for more info."""
        event: guild_events.GuildUpdateEvent | None
        if self._enabled_for_event(guild_events.GuildUpdateEvent):
            guild_id = snowflakes.Snowflake(payload["id"])
            old = self._cache.get_guild(guild_id) if self._cache else None
            event = self._event_factory.deserialize_guild_update_event(shard, payload, old_guild=old)

            # We also filter here to prevent iterating over them and calling a function that won't do anything
            emojis = event.emojis if self._cache_enabled_for(config.CacheComponents.EMOJIS) else None
            stickers = event.stickers if self._cache_enabled_for(config.CacheComponents.GUILD_STICKERS) else None
            guild = event.guild if self._cache_enabled_for(config.CacheComponents.GUILDS) else None
            roles = event.roles if self._cache_enabled_for(config.CacheComponents.ROLES) else None

        elif self._cache:
            _LOGGER.log(ux.TRACE, "Skipping on_guild_update raw dispatch due to lack of any registered listeners")
            event = None

            gd = self._entity_factory.deserialize_gateway_guild(payload, user_id=shard.get_user_id())
            emojis = gd.emojis() if self._cache_enabled_for(config.CacheComponents.EMOJIS) else None
            stickers = gd.stickers() if self._cache_enabled_for(config.CacheComponents.GUILD_STICKERS) else None
            guild = gd.guild() if self._cache_enabled_for(config.CacheComponents.GUILDS) else None
            guild_id = gd.id
            roles = gd.roles() if self._cache_enabled_for(config.CacheComponents.ROLES) else None

        else:
            _LOGGER.log(
                ux.TRACE, "Skipping on_guild_update raw dispatch due to lack of any registered listeners or cache need"
            )
            return

        if self._cache:
            if guild:
                self._cache.update_guild(guild)

            if emojis:
                self._cache.clear_emojis_for_guild(guild_id)
                for emoji in emojis.values():
                    self._cache.set_emoji(emoji)

            if stickers:
                self._cache.clear_stickers_for_guild(guild_id)
                for sticker in stickers.values():
                    self._cache.set_sticker(sticker)

            if roles:
                self._cache.clear_roles_for_guild(guild_id)
                for role in roles.values():
                    self._cache.set_role(role)

        if event:
            self.dispatch(event)

    @event_manager_base.filtered(
        (guild_events.GuildLeaveEvent, guild_events.GuildUnavailableEvent),
        config.CacheComponents.GUILDS
        | config.CacheComponents.GUILD_CHANNELS
        | config.CacheComponents.EMOJIS
        | config.CacheComponents.GUILD_STICKERS
        | config.CacheComponents.ROLES
        | config.CacheComponents.PRESENCES
        | config.CacheComponents.VOICE_STATES
        | config.CacheComponents.MEMBERS
        | config.CacheComponents.GUILD_THREADS,
    )
    def on_guild_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-delete for more info."""
        event: guild_events.GuildUnavailableEvent | guild_events.GuildLeaveEvent
        if payload.get("unavailable"):
            event = self._event_factory.deserialize_guild_unavailable_event(shard, payload)

            if self._cache:
                self._cache.set_guild_availability(event.guild_id, False)

        else:
            old: guilds.GatewayGuild | None = None
            if self._cache:
                guild_id = snowflakes.Snowflake(payload["id"])
                #  TODO: this doesn't work in all intent scenarios
                old = self._cache.delete_guild(guild_id)
                self._cache.clear_voice_states_for_guild(guild_id)
                self._cache.clear_invites_for_guild(guild_id)
                self._cache.clear_members_for_guild(guild_id)
                self._cache.clear_presences_for_guild(guild_id)
                self._cache.clear_guild_channels_for_guild(guild_id)
                self._cache.clear_threads_for_guild(guild_id)
                self._cache.clear_emojis_for_guild(guild_id)
                self._cache.clear_stickers_for_guild(guild_id)
                self._cache.clear_roles_for_guild(guild_id)

            event = self._event_factory.deserialize_guild_leave_event(shard, payload, old_guild=old)

        self.dispatch(event)

    @event_manager_base.filtered(guild_events.BanCreateEvent)
    def on_guild_ban_add(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-ban-add for more info."""
        self.dispatch(self._event_factory.deserialize_guild_ban_add_event(shard, payload))

    @event_manager_base.filtered(guild_events.BanDeleteEvent)
    def on_guild_ban_remove(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-ban-remove for more info."""
        self.dispatch(self._event_factory.deserialize_guild_ban_remove_event(shard, payload))

    @event_manager_base.filtered(guild_events.EmojisUpdateEvent, config.CacheComponents.EMOJIS)
    def on_guild_emojis_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-emojis-update for more info."""
        guild_id = snowflakes.Snowflake(payload["guild_id"])
        old = list(self._cache.clear_emojis_for_guild(guild_id).values()) if self._cache else None

        event = self._event_factory.deserialize_guild_emojis_update_event(shard, payload, old_emojis=old)

        if self._cache:
            for emoji in event.emojis:
                self._cache.set_emoji(emoji)

        self.dispatch(event)

    @event_manager_base.filtered(guild_events.StickersUpdateEvent, config.CacheComponents.EMOJIS)
    def on_guild_stickers_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-stickers-update for more info."""
        guild_id = snowflakes.Snowflake(payload["guild_id"])
        old = list(self._cache.clear_stickers_for_guild(guild_id).values()) if self._cache else None

        event = self._event_factory.deserialize_guild_stickers_update_event(shard, payload, old_stickers=old)

        if self._cache:
            for sticker in event.stickers:
                self._cache.set_sticker(sticker)

        self.dispatch(event)

    @event_manager_base.filtered(())  # An empty sequence here means that this method will always be skipped.
    def on_guild_integrations_update(self, _: gateway_shard.GatewayShard, __: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-integrations-update for more info."""
        # This is only here to stop this being logged or dispatched as an "unknown event".
        # This event is made redundant by INTEGRATION_CREATE/DELETE/UPDATE and is thus not parsed or dispatched.
        raise NotImplementedError

    @event_manager_base.filtered(guild_events.IntegrationCreateEvent)
    def on_integration_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        event = self._event_factory.deserialize_integration_create_event(shard, payload)
        self.dispatch(event)

    @event_manager_base.filtered(guild_events.IntegrationUpdateEvent)
    def on_integration_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        event = self._event_factory.deserialize_integration_update_event(shard, payload)
        self.dispatch(event)

    @event_manager_base.filtered(guild_events.IntegrationDeleteEvent)
    def on_integration_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        event = self._event_factory.deserialize_integration_delete_event(shard, payload)
        self.dispatch(event)

    @event_manager_base.filtered(member_events.MemberCreateEvent, config.CacheComponents.MEMBERS)
    def on_guild_member_add(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-member-add for more info."""
        event = self._event_factory.deserialize_guild_member_add_event(shard, payload)

        if self._cache:
            self._cache.update_member(event.member)

        self.dispatch(event)

    @event_manager_base.filtered(member_events.MemberDeleteEvent, config.CacheComponents.MEMBERS)
    def on_guild_member_remove(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-member-remove for more info."""
        old: guilds.Member | None = None
        if self._cache:
            old = self._cache.delete_member(
                snowflakes.Snowflake(payload["guild_id"]), snowflakes.Snowflake(payload["user"]["id"])
            )

        event = self._event_factory.deserialize_guild_member_remove_event(shard, payload, old_member=old)
        self.dispatch(event)

    @event_manager_base.filtered(member_events.MemberUpdateEvent, config.CacheComponents.MEMBERS)
    def on_guild_member_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-member-update for more info."""
        old: guilds.Member | None = None
        if self._cache:
            old = self._cache.get_member(
                snowflakes.Snowflake(payload["guild_id"]), snowflakes.Snowflake(payload["user"]["id"])
            )

        event = self._event_factory.deserialize_guild_member_update_event(shard, payload, old_member=old)

        if self._cache:
            self._cache.update_member(event.member)

        self.dispatch(event)

    @event_manager_base.filtered(shard_events.MemberChunkEvent, config.CacheComponents.MEMBERS)
    def on_guild_members_chunk(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-members-chunk for more info."""
        event = self._event_factory.deserialize_guild_member_chunk_event(shard, payload)

        if self._cache:
            for member in event.members.values():
                self._cache.set_member(member)

            for presence in event.presences.values():
                self._cache.set_presence(presence)

        self.dispatch(event)

    @event_manager_base.filtered(role_events.RoleCreateEvent, config.CacheComponents.ROLES)
    def on_guild_role_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-role-create for more info."""
        event = self._event_factory.deserialize_guild_role_create_event(shard, payload)

        if self._cache:
            self._cache.set_role(event.role)

        self.dispatch(event)

    @event_manager_base.filtered(role_events.RoleUpdateEvent, config.CacheComponents.ROLES)
    def on_guild_role_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-role-update for more info."""
        old = self._cache.get_role(snowflakes.Snowflake(payload["role"]["id"])) if self._cache else None
        event = self._event_factory.deserialize_guild_role_update_event(shard, payload, old_role=old)

        if self._cache:
            self._cache.update_role(event.role)

        self.dispatch(event)

    @event_manager_base.filtered(role_events.RoleDeleteEvent, config.CacheComponents.ROLES)
    def on_guild_role_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-role-delete for more info."""
        old: guilds.Role | None = None
        if self._cache:
            old = self._cache.delete_role(snowflakes.Snowflake(payload["role_id"]))

        event = self._event_factory.deserialize_guild_role_delete_event(shard, payload, old_role=old)

        self.dispatch(event)

    @event_manager_base.filtered(channel_events.InviteCreateEvent, config.CacheComponents.INVITES)
    def on_invite_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#invite-create for more info."""
        event = self._event_factory.deserialize_invite_create_event(shard, payload)

        if self._cache:
            self._cache.set_invite(event.invite)

        self.dispatch(event)

    @event_manager_base.filtered(channel_events.InviteDeleteEvent, config.CacheComponents.INVITES)
    def on_invite_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#invite-delete for more info."""
        old: invites.InviteWithMetadata | None = None
        if self._cache:
            old = self._cache.delete_invite(payload["code"])

        event = self._event_factory.deserialize_invite_delete_event(shard, payload, old_invite=old)

        self.dispatch(event)

    @event_manager_base.filtered(
        (message_events.GuildMessageCreateEvent, message_events.DMMessageCreateEvent), config.CacheComponents.MESSAGES
    )
    def on_message_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-create for more info."""
        event = self._event_factory.deserialize_message_create_event(shard, payload)

        if self._cache:
            self._cache.set_message(event.message)

        self.dispatch(event)

    @event_manager_base.filtered(
        (message_events.GuildMessageUpdateEvent, message_events.DMMessageUpdateEvent), config.CacheComponents.MESSAGES
    )
    def on_message_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-update for more info."""
        old = self._cache.get_message(snowflakes.Snowflake(payload["id"])) if self._cache else None
        event = self._event_factory.deserialize_message_update_event(shard, payload, old_message=old)

        if self._cache:
            self._cache.update_message(event.message)

        self.dispatch(event)

    @event_manager_base.filtered(
        (message_events.GuildMessageDeleteEvent, message_events.DMMessageDeleteEvent), config.CacheComponents.MESSAGES
    )
    def on_message_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-delete for more info."""
        if self._cache:
            message_id = snowflakes.Snowflake(payload["id"])
            old_message = self._cache.delete_message(message_id)
        else:
            old_message = None

        event = self._event_factory.deserialize_message_delete_event(shard, payload, old_message=old_message)

        self.dispatch(event)

    @event_manager_base.filtered(
        (message_events.GuildMessageDeleteEvent, message_events.DMMessageDeleteEvent), config.CacheComponents.MESSAGES
    )
    def on_message_delete_bulk(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-delete-bulk for more info."""
        old_messages = {}

        if self._cache:
            for raw_message_id in payload["ids"]:
                message_id = snowflakes.Snowflake(raw_message_id)

                if message := self._cache.delete_message(message_id):
                    old_messages[message_id] = message

        self.dispatch(
            self._event_factory.deserialize_guild_message_delete_bulk_event(shard, payload, old_messages=old_messages)
        )

    @event_manager_base.filtered((reaction_events.GuildReactionAddEvent, reaction_events.DMReactionAddEvent))
    def on_message_reaction_add(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-reaction-add for more info."""
        self.dispatch(self._event_factory.deserialize_message_reaction_add_event(shard, payload))

    # TODO: this is unlikely but reaction cache?

    @event_manager_base.filtered((reaction_events.GuildReactionDeleteEvent, reaction_events.DMReactionDeleteEvent))
    def on_message_reaction_remove(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove for more info."""
        self.dispatch(self._event_factory.deserialize_message_reaction_remove_event(shard, payload))

    @event_manager_base.filtered(
        (reaction_events.GuildReactionDeleteAllEvent, reaction_events.DMReactionDeleteAllEvent)
    )
    def on_message_reaction_remove_all(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove-all for more info."""
        self.dispatch(self._event_factory.deserialize_message_reaction_remove_all_event(shard, payload))

    @event_manager_base.filtered(
        (reaction_events.GuildReactionDeleteEmojiEvent, reaction_events.DMReactionDeleteEmojiEvent)
    )
    def on_message_reaction_remove_emoji(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove-emoji for more info."""
        self.dispatch(self._event_factory.deserialize_message_reaction_remove_emoji_event(shard, payload))

    @event_manager_base.filtered(guild_events.PresenceUpdateEvent, config.CacheComponents.PRESENCES)
    def on_presence_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#presence-update for more info."""
        old: presences_.MemberPresence | None = None

        if self._cache:
            old = self._cache.get_presence(
                snowflakes.Snowflake(payload["guild_id"]), snowflakes.Snowflake(payload["user"]["id"])
            )

        event = self._event_factory.deserialize_presence_update_event(shard, payload, old_presence=old)

        if self._cache and event.presence.visible_status is presences_.Status.OFFLINE:
            self._cache.delete_presence(event.presence.guild_id, event.presence.user_id)
        elif self._cache:
            self._cache.update_presence(event.presence)

        # TODO: update user here when partial_user is set self._cache.update_user(event.partial_user)
        self.dispatch(event)

    @event_manager_base.filtered((typing_events.GuildTypingEvent, typing_events.DMTypingEvent))
    def on_typing_start(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#typing-start for more info."""
        self.dispatch(self._event_factory.deserialize_typing_start_event(shard, payload))

    @event_manager_base.filtered(user_events.OwnUserUpdateEvent, config.CacheComponents.ME)
    def on_user_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#user-update for more info."""
        old = self._cache.get_me() if self._cache else None
        event = self._event_factory.deserialize_own_user_update_event(shard, payload, old_user=old)

        if self._cache:
            self._cache.update_me(event.user)

        self.dispatch(event)

    @event_manager_base.filtered(voice_events.VoiceStateUpdateEvent, config.CacheComponents.VOICE_STATES)
    def on_voice_state_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#voice-state-update for more info."""
        old: voices.VoiceState | None = None
        if self._cache:
            old = self._cache.get_voice_state(
                snowflakes.Snowflake(payload["guild_id"]), snowflakes.Snowflake(payload["user_id"])
            )

        event = self._event_factory.deserialize_voice_state_update_event(shard, payload, old_state=old)

        if self._cache and event.state.channel_id is None:
            self._cache.delete_voice_state(event.state.guild_id, event.state.user_id)
        elif self._cache:
            self._cache.update_voice_state(event.state)

        self.dispatch(event)

    @event_manager_base.filtered(voice_events.VoiceServerUpdateEvent)
    def on_voice_server_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#voice-server-update for more info."""
        self.dispatch(self._event_factory.deserialize_voice_server_update_event(shard, payload))

    @event_manager_base.filtered(channel_events.WebhookUpdateEvent)
    def on_webhooks_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#webhooks-update for more info."""
        self.dispatch(self._event_factory.deserialize_webhook_update_event(shard, payload))

    @event_manager_base.filtered(
        (
            interaction_events.CommandInteractionCreateEvent,
            interaction_events.ComponentInteractionCreateEvent,
            interaction_events.AutocompleteInteractionCreateEvent,
            interaction_events.ModalInteractionCreateEvent,
        )
    )
    def on_interaction_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#interaction-create for more info."""
        self.dispatch(self._event_factory.deserialize_interaction_create_event(shard, payload))

    @event_manager_base.filtered(scheduled_events.ScheduledEventCreateEvent)
    def on_guild_scheduled_event_create(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-create for more info."""
        self.dispatch(self._event_factory.deserialize_scheduled_event_create_event(shard, payload))

    @event_manager_base.filtered(scheduled_events.ScheduledEventDeleteEvent)
    def on_guild_scheduled_event_delete(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-delete for more info."""
        self.dispatch(self._event_factory.deserialize_scheduled_event_delete_event(shard, payload))

    @event_manager_base.filtered(scheduled_events.ScheduledEventUpdateEvent)
    def on_guild_scheduled_event_update(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-update for more info."""
        self.dispatch(self._event_factory.deserialize_scheduled_event_update_event(shard, payload))

    @event_manager_base.filtered(scheduled_events.ScheduledEventUserAddEvent)
    def on_guild_scheduled_event_user_add(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-user-add for more info."""  # noqa: E501
        self.dispatch(self._event_factory.deserialize_scheduled_event_user_add_event(shard, payload))

    @event_manager_base.filtered(scheduled_events.ScheduledEventUserRemoveEvent)
    def on_guild_scheduled_event_user_remove(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-user-remove for more info."""  # noqa: E501
        self.dispatch(self._event_factory.deserialize_scheduled_event_user_remove_event(shard, payload))

    @event_manager_base.filtered(guild_events.AuditLogEntryCreateEvent)
    def on_guild_audit_log_entry_create(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#guild-audit-log-entry-create for more info."""
        self.dispatch(self._event_factory.deserialize_audit_log_entry_create_event(shard, payload))

    @event_manager_base.filtered(monetization_events.EntitlementCreateEvent)
    def on_entitlement_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#entitlement-create for more info."""
        self.dispatch(self._event_factory.deserialize_entitlement_create_event(shard, payload))

    @event_manager_base.filtered(monetization_events.EntitlementDeleteEvent)
    def on_entitlement_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#entitlement-delete for more info."""
        self.dispatch(self._event_factory.deserialize_entitlement_delete_event(shard, payload))

    @event_manager_base.filtered(monetization_events.EntitlementUpdateEvent)
    def on_entitlement_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#entitlement-update for more info."""
        self.dispatch(self._event_factory.deserialize_entitlement_update_event(shard, payload))

    @event_manager_base.filtered(stage_events.StageInstanceCreateEvent)
    def on_stage_instance_create(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        self.dispatch(self._event_factory.deserialize_stage_instance_create_event(shard, payload))

    @event_manager_base.filtered(stage_events.StageInstanceUpdateEvent)
    def on_stage_instance_update(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        self.dispatch(self._event_factory.deserialize_stage_instance_update_event(shard, payload))

    @event_manager_base.filtered(stage_events.StageInstanceDeleteEvent)
    def on_stage_instance_delete(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        self.dispatch(self._event_factory.deserialize_stage_instance_delete_event(shard, payload))

    @event_manager_base.filtered(poll_events.PollVoteCreateEvent)
    def on_message_poll_vote_add(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-poll-vote-add for more info."""
        self.dispatch(self._event_factory.deserialize_poll_vote_create_event(shard, payload))

    @event_manager_base.filtered(poll_events.PollVoteDeleteEvent)
    def on_message_poll_vote_remove(self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject) -> None:
        """See https://discord.com/developers/docs/topics/gateway-events#message-poll-vote-remove for more info."""
        self.dispatch(self._event_factory.deserialize_poll_vote_delete_event(shard, payload))

    @event_manager_base.filtered(auto_mod_events.AutoModRuleCreateEvent)
    def on_auto_moderation_rule_create(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway#auto-moderation-rule-create for more info."""
        self.dispatch(self._event_factory.deserialize_auto_mod_rule_create_event(shard, payload))

    @event_manager_base.filtered(auto_mod_events.AutoModRuleUpdateEvent)
    def on_auto_moderation_rule_update(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway#auto-moderation-rule-update for more info."""
        self.dispatch(self._event_factory.deserialize_auto_mod_rule_update_event(shard, payload))

    @event_manager_base.filtered(auto_mod_events.AutoModRuleDeleteEvent)
    def on_auto_moderation_rule_delete(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway#auto-moderation-rule-delete for more info."""
        self.dispatch(self._event_factory.deserialize_auto_mod_rule_delete_event(shard, payload))

    @event_manager_base.filtered(auto_mod_events.AutoModActionExecutionEvent)
    def on_auto_moderation_action_execution(
        self, shard: gateway_shard.GatewayShard, payload: data_binding.JSONObject
    ) -> None:
        """See https://discord.com/developers/docs/topics/gateway#auto-moderation-action-execution for more info."""
        self.dispatch(self._event_factory.deserialize_auto_mod_action_execution_event(shard, payload))



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import concurrent.futures
import contextlib
import sys

import mock
import pytest

from hikari import errors
from hikari.impl import config
from hikari.impl import entity_factory as entity_factory_impl
from hikari.impl import interaction_server as interaction_server_impl
from hikari.impl import rest as rest_impl
from hikari.impl import rest_bot as rest_bot_impl
from hikari.internal import aio
from hikari.internal import signals
from hikari.internal import ux
from tests.hikari import hikari_test_helpers


class TestRESTBot:
    @pytest.fixture
    def mock_interaction_server(self):
        return mock.Mock(interaction_server_impl.InteractionServer)

    @pytest.fixture
    def mock_rest_client(self):
        return mock.Mock(rest_impl.RESTClientImpl)

    @pytest.fixture
    def mock_entity_factory(self):
        return mock.Mock(entity_factory_impl.EntityFactoryImpl)

    @pytest.fixture
    def mock_http_settings(self):
        return mock.Mock(config.HTTPSettings)

    @pytest.fixture
    def mock_proxy_settings(self):
        return mock.Mock(config.ProxySettings)

    @pytest.fixture
    def mock_executor(self):
        return mock.Mock(concurrent.futures.Executor)

    @pytest.fixture
    def mock_rest_bot(
        self,
        mock_interaction_server,
        mock_rest_client,
        mock_entity_factory,
        mock_http_settings,
        mock_proxy_settings,
        mock_executor,
    ):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(ux, "init_logging"))
        stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))
        stack.enter_context(mock.patch.object(rest_bot_impl.RESTBot, "print_banner"))
        stack.enter_context(
            mock.patch.object(entity_factory_impl, "EntityFactoryImpl", return_value=mock_entity_factory)
        )
        stack.enter_context(mock.patch.object(rest_impl, "RESTClientImpl", return_value=mock_rest_client))
        stack.enter_context(
            mock.patch.object(interaction_server_impl, "InteractionServer", return_value=mock_interaction_server)
        )

        with stack:
            return hikari_test_helpers.mock_class_namespace(rest_bot_impl.RESTBot, slots_=False)(
                "token",
                http_settings=mock_http_settings,
                proxy_settings=mock_proxy_settings,
                executor=mock_executor,
                max_retries=0,
            )

    def test___init__(
        self, mock_http_settings, mock_proxy_settings, mock_entity_factory, mock_rest_client, mock_interaction_server
    ):
        mock_executor = mock.Mock()

        stack = contextlib.ExitStack()
        patched_init_logging = stack.enter_context(mock.patch.object(ux, "init_logging"))
        patched_warn_if_not_optimized = stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))
        patched_print_banner = stack.enter_context(mock.patch.object(rest_bot_impl.RESTBot, "print_banner"))
        patched_entity_factory = stack.enter_context(
            mock.patch.object(entity_factory_impl, "EntityFactoryImpl", return_value=mock_entity_factory)
        )
        patched_rest_client = stack.enter_context(
            mock.patch.object(rest_impl, "RESTClientImpl", return_value=mock_rest_client)
        )
        patched_interaction_server = stack.enter_context(
            mock.patch.object(interaction_server_impl, "InteractionServer", return_value=mock_interaction_server)
        )

        with stack:
            result = rest_bot_impl.RESTBot(
                "token",
                "token_type",
                b"2123123123123132",
                allow_color=False,
                banner="a banner",
                suppress_optimization_warning=True,
                executor=mock_executor,
                force_color=True,
                http_settings=mock_http_settings,
                logs="ERROR",
                max_rate_limit=32123123,
                max_retries=0,
                proxy_settings=mock_proxy_settings,
                rest_url="hresresres",
            )

        patched_init_logging.assert_called_once_with("ERROR", allow_color=False, force_color=True)
        patched_warn_if_not_optimized.assert_called_once_with(suppress=True)
        patched_entity_factory.assert_called_once_with(result)
        patched_print_banner.assert_called_once_with("a banner", allow_color=False, force_color=True)
        patched_rest_client.assert_called_once_with(
            cache=None,
            entity_factory=mock_entity_factory,
            executor=mock_executor,
            http_settings=mock_http_settings,
            max_rate_limit=32123123,
            max_retries=0,
            proxy_settings=mock_proxy_settings,
            rest_url="hresresres",
            token="token",
            token_type="token_type",
        )
        patched_interaction_server.assert_called_once_with(
            entity_factory=mock_entity_factory, public_key=b"2123123123123132", rest_client=mock_rest_client
        )
        assert result.interaction_server is mock_interaction_server
        assert result.rest is mock_rest_client
        assert result.entity_factory is mock_entity_factory
        assert result.http_settings is mock_http_settings
        assert result.proxy_settings is mock_proxy_settings
        assert result.executor is mock_executor

    def test___init___parses_string_public_key(self):
        cls = hikari_test_helpers.mock_class_namespace(rest_bot_impl.RESTBot, print_banner=mock.Mock())

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(ux, "init_logging"))
        stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))
        stack.enter_context(mock.patch.object(rest_bot_impl.RESTBot, "print_banner"))
        stack.enter_context(mock.patch.object(interaction_server_impl, "InteractionServer"))

        with stack:
            result = cls(object(), "token_type", "6f66646f646f646f6f")

            interaction_server_impl.InteractionServer.assert_called_once_with(
                entity_factory=result.entity_factory, public_key=b"ofdododoo", rest_client=result.rest
            )

    def test___init___strips_token(self):
        cls = hikari_test_helpers.mock_class_namespace(rest_bot_impl.RESTBot, print_banner=mock.Mock())

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(ux, "init_logging"))
        stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))
        rest_client = stack.enter_context(mock.patch.object(rest_impl, "RESTClientImpl"))
        http_settings = stack.enter_context(mock.patch.object(config, "HTTPSettings"))
        proxy_settings = stack.enter_context(mock.patch.object(config, "ProxySettings"))
        stack.enter_context(mock.patch.object(interaction_server_impl, "InteractionServer"))

        with stack:
            result = cls("\n\r sddsa tokenoken \n", "token_type")

        rest_client.assert_called_once_with(
            cache=None,
            entity_factory=result.entity_factory,
            executor=None,
            http_settings=http_settings.return_value,
            max_rate_limit=300.0,
            max_retries=3,
            proxy_settings=proxy_settings.return_value,
            rest_url=None,
            token="sddsa tokenoken",
            token_type="token_type",
        )

    def test___init___generates_default_settings(self):
        cls = hikari_test_helpers.mock_class_namespace(rest_bot_impl.RESTBot, print_banner=mock.Mock())
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(ux, "init_logging"))
        stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))
        stack.enter_context(mock.patch.object(rest_bot_impl.RESTBot, "print_banner"))
        stack.enter_context(mock.patch.object(rest_impl, "RESTClientImpl"))
        stack.enter_context(mock.patch.object(config, "HTTPSettings"))
        stack.enter_context(mock.patch.object(config, "ProxySettings"))
        stack.enter_context(mock.patch.object(interaction_server_impl, "InteractionServer"))

        with stack:
            result = cls("token")

            rest_impl.RESTClientImpl.assert_called_once_with(
                cache=None,
                entity_factory=result.entity_factory,
                executor=None,
                http_settings=config.HTTPSettings.return_value,
                max_rate_limit=300.0,
                max_retries=3,
                proxy_settings=config.ProxySettings.return_value,
                rest_url=None,
                token="token",
                token_type="Bot",
            )

            config.HTTPSettings.assert_called_once()
            config.ProxySettings.assert_called_once()
            assert result.http_settings is config.HTTPSettings.return_value
            assert result.proxy_settings is config.ProxySettings.return_value

    @pytest.mark.parametrize(("close_event", "expected"), [(object(), True), (None, False)])
    def test_is_alive_property(self, mock_rest_bot, close_event, expected):
        mock_rest_bot._close_event = close_event
        assert mock_rest_bot.is_alive is expected

    def test_print_banner(self, mock_rest_bot):
        with mock.patch.object(ux, "print_banner") as print_banner:
            mock_rest_bot.print_banner(
                "okokok", allow_color=True, force_color=False, extra_args={"test_key": "test_value"}
            )

        print_banner.assert_called_once_with(
            "okokok", allow_color=True, force_color=False, extra_args={"test_key": "test_value"}
        )

    def test_add_shutdown_callback(self, mock_rest_bot: rest_bot_impl.RESTBot):
        callback = mock.Mock()
        mock_rest_bot.add_shutdown_callback(callback)

        assert callback in mock_rest_bot.on_shutdown

    def test_remove_shutdown_callback(self, mock_rest_bot: rest_bot_impl.RESTBot):
        callback = mock.Mock()
        mock_rest_bot.add_shutdown_callback(callback)

        mock_rest_bot.remove_shutdown_callback(callback)

        assert callback not in mock_rest_bot.on_shutdown

    def test_remove_shutdown_callback_when_not_present(self, mock_rest_bot: rest_bot_impl.RESTBot):
        callback = mock.Mock()

        with pytest.raises(ValueError, match=".*"):
            mock_rest_bot.remove_shutdown_callback(callback)

    def test_add_startup_callback(self, mock_rest_bot: rest_bot_impl.RESTBot):
        callback = mock.Mock()
        mock_rest_bot.add_startup_callback(callback)

        assert callback in mock_rest_bot.on_startup

    def test_remove_startup_callback(self, mock_rest_bot: rest_bot_impl.RESTBot):
        callback = mock.Mock()
        mock_rest_bot.add_startup_callback(callback)

        mock_rest_bot.remove_startup_callback(callback)

        assert callback not in mock_rest_bot.on_startup

    def test_remove_startup_callback_when_not_present(self, mock_rest_bot: rest_bot_impl.RESTBot):
        callback = mock.Mock()

        with pytest.raises(ValueError, match=".*"):
            mock_rest_bot.remove_startup_callback(callback)

    @pytest.mark.asyncio
    async def test_close(
        self, mock_rest_bot: rest_bot_impl.RESTBot, mock_interaction_server: mock.Mock, mock_rest_client: mock.Mock
    ):
        mock_shutdown_1 = mock.AsyncMock()
        mock_shutdown_2 = mock.AsyncMock()
        mock_rest_bot._close_event = close_event = mock.Mock()
        mock_interaction_server.close = mock.AsyncMock()
        mock_rest_bot._is_closing = False
        mock_rest_bot.add_shutdown_callback(mock_shutdown_1)
        mock_rest_bot.add_shutdown_callback(mock_shutdown_2)

        await mock_rest_bot.close()

        mock_interaction_server.close.assert_awaited_once()
        mock_rest_client.close.assert_awaited_once()
        close_event.set.assert_called_once()
        assert mock_rest_bot._is_closing is False
        mock_shutdown_1.assert_awaited_once_with(mock_rest_bot)
        mock_shutdown_2.assert_awaited_once_with(mock_rest_bot)

    @pytest.mark.asyncio
    async def test_close_when_shutdown_callback_raises(
        self, mock_rest_bot: rest_bot_impl.RESTBot, mock_interaction_server: mock.Mock, mock_rest_client: mock.Mock
    ):
        mock_error = KeyError("Too many catgirls")
        mock_shutdown_1 = mock.AsyncMock(side_effect=mock_error)
        mock_shutdown_2 = mock.AsyncMock()
        mock_rest_bot._close_event = close_event = mock.Mock()
        mock_interaction_server.close = mock.AsyncMock()
        mock_rest_bot._is_closing = False
        mock_rest_bot.add_shutdown_callback(mock_shutdown_1)
        mock_rest_bot.add_shutdown_callback(mock_shutdown_2)

        with pytest.raises(KeyError) as exc_info:
            await mock_rest_bot.close()

        assert exc_info.value is mock_error
        mock_interaction_server.close.assert_awaited_once()
        mock_rest_client.close.assert_awaited_once()
        close_event.set.assert_called_once()
        assert mock_rest_bot._is_closing is False
        mock_shutdown_1.assert_awaited_once_with(mock_rest_bot)
        mock_shutdown_2.assert_not_called()

    @pytest.mark.asyncio
    async def test_close_when_is_closing(
        self, mock_rest_bot: rest_bot_impl.RESTBot, mock_interaction_server: mock.Mock, mock_rest_client: mock.Mock
    ):
        mock_shutdown_1 = mock.AsyncMock()
        mock_shutdown_2 = mock.AsyncMock()
        mock_rest_bot._close_event = mock.Mock()
        mock_interaction_server.close = mock.AsyncMock()
        mock_rest_bot._is_closing = True
        mock_rest_bot.join = mock.AsyncMock()
        mock_rest_bot.add_shutdown_callback(mock_shutdown_1)
        mock_rest_bot.add_shutdown_callback(mock_shutdown_2)

        await mock_rest_bot.close()

        mock_interaction_server.close.assert_not_called()
        mock_rest_client.close.assert_not_called()
        mock_rest_bot._close_event.set.assert_not_called()
        mock_rest_bot.join.assert_awaited_once()
        assert mock_rest_bot._is_closing is True
        mock_shutdown_1.assert_not_called()
        mock_shutdown_2.assert_not_called()

    @pytest.mark.asyncio
    async def test_close_when_inactive(self, mock_rest_bot):
        with pytest.raises(errors.ComponentStateConflictError):
            await mock_rest_bot.close()

    @pytest.mark.asyncio
    async def test_join(self, mock_rest_bot):
        mock_rest_bot._close_event = mock.AsyncMock()

        await mock_rest_bot.join()

        mock_rest_bot._close_event.wait.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_join_when_not_alive(self, mock_rest_bot):
        with pytest.raises(errors.ComponentStateConflictError):
            await mock_rest_bot.join()

    @pytest.mark.asyncio
    async def test_on_interaction(self, mock_rest_bot, mock_interaction_server):
        mock_interaction_server.on_interaction = mock.AsyncMock()

        result = await mock_rest_bot.on_interaction(b"1", b"2", b"3")

        assert result is mock_interaction_server.on_interaction.return_value
        mock_interaction_server.on_interaction.assert_awaited_once_with(b"1", b"2", b"3")

    def test_run(self, mock_rest_bot):
        mock_socket = object()
        mock_context = object()
        mock_rest_bot._executor = None
        mock_rest_bot.start = mock.Mock()
        mock_rest_bot.join = mock.Mock()

        stack = contextlib.ExitStack()
        check_for_updates = stack.enter_context(mock.patch.object(ux, "check_for_updates"))
        handle_interrupts = stack.enter_context(
            mock.patch.object(signals, "handle_interrupts", return_value=hikari_test_helpers.ContextManagerMock())
        )
        get_or_make_loop = stack.enter_context(mock.patch.object(aio, "get_or_make_loop"))

        with stack:
            mock_rest_bot.run(
                asyncio_debug=False,
                backlog=321,
                check_for_updates=False,
                close_loop=False,
                close_passed_executor=False,
                coroutine_tracking_depth=32123,
                enable_signal_handlers=True,
                propagate_interrupts=True,
                host="192.168.1.102",
                path="pathathath",
                port=4554,
                reuse_address=True,
                reuse_port=False,
                shutdown_timeout=534.534,
                socket=mock_socket,
                ssl_context=mock_context,
            )

        check_for_updates.assert_not_called()
        handle_interrupts.assert_called_once_with(
            enabled=True, loop=get_or_make_loop.return_value, propagate_interrupts=True
        )
        handle_interrupts.return_value.assert_used_once()

        mock_rest_bot.start.assert_called_once_with(
            backlog=321,
            check_for_updates=False,
            host="192.168.1.102",
            path="pathathath",
            port=4554,
            reuse_address=True,
            reuse_port=False,
            socket=mock_socket,
            shutdown_timeout=534.534,
            ssl_context=mock_context,
        )
        mock_rest_bot.join.assert_called_once_with()

        assert get_or_make_loop.return_value.run_until_complete.call_count == 2
        get_or_make_loop.return_value.run_until_complete.assert_has_calls(
            [mock.call(mock_rest_bot.start.return_value), mock.call(mock_rest_bot.join.return_value)]
        )
        get_or_make_loop.return_value.close.assert_not_called()

    def test_run_when_close_loop(self, mock_rest_bot):
        mock_rest_bot.start = mock.Mock()
        mock_rest_bot.join = mock.Mock()

        with mock.patch.object(aio, "get_or_make_loop") as get_or_make_loop:
            with mock.patch.object(aio, "destroy_loop") as destroy_loop:
                mock_rest_bot.run(close_loop=True)

        destroy_loop.assert_called_once_with(get_or_make_loop.return_value, rest_bot_impl._LOGGER)

    def test_run_when_asyncio_debug(self, mock_rest_bot):
        mock_rest_bot.start = mock.Mock()
        mock_rest_bot.join = mock.Mock()

        with mock.patch.object(aio, "get_or_make_loop") as get_or_make_loop:
            mock_rest_bot.run(asyncio_debug=True, close_loop=False)

        get_or_make_loop.return_value.set_debug.assert_called_once_with(True)

    def test_run_with_coroutine_tracking_depth(self, mock_rest_bot):
        mock_rest_bot.start = mock.Mock()
        mock_rest_bot.join = mock.Mock()

        with mock.patch.object(aio, "get_or_make_loop"):
            with mock.patch.object(
                sys, "set_coroutine_origin_tracking_depth", side_effect=AttributeError, create=True
            ) as set_tracking_depth:
                mock_rest_bot.run(coroutine_tracking_depth=42, close_loop=False)

        set_tracking_depth.assert_called_once_with(42)

    def test_run_when_already_running(self, mock_rest_bot):
        mock_rest_bot._close_event = object()

        with pytest.raises(errors.ComponentStateConflictError):
            mock_rest_bot.run()

    def test_run_closes_executor_when_present(self, mock_rest_bot, mock_executor):
        mock_rest_bot.start = mock.Mock()
        mock_rest_bot.join = mock.Mock()

        with mock.patch.object(aio, "get_or_make_loop"):
            mock_rest_bot.run(
                asyncio_debug=False,
                backlog=321,
                check_for_updates=False,
                close_loop=False,
                close_passed_executor=True,
                coroutine_tracking_depth=32123,
                enable_signal_handlers=True,
                host="192.168.1.102",
                path="pathathath",
                port=4554,
                reuse_address=True,
                reuse_port=False,
                shutdown_timeout=534.534,
                socket=object(),
                ssl_context=object(),
            )

        mock_executor.shutdown.assert_called_once_with(wait=True)
        assert mock_rest_bot.executor is None

    def test_run_ignores_close_executor_when_not_present(self, mock_rest_bot):
        mock_rest_bot.start = mock.Mock()
        mock_rest_bot.join = mock.Mock()
        mock_rest_bot._executor = None

        with mock.patch.object(aio, "get_or_make_loop"):
            mock_rest_bot.run(
                asyncio_debug=False,
                backlog=321,
                check_for_updates=False,
                close_loop=False,
                close_passed_executor=True,
                coroutine_tracking_depth=32123,
                enable_signal_handlers=True,
                host="192.168.1.102",
                path="pathathath",
                port=4554,
                reuse_address=True,
                reuse_port=False,
                shutdown_timeout=534.534,
                socket=object(),
                ssl_context=object(),
            )

        assert mock_rest_bot.executor is None

    @pytest.mark.asyncio
    async def test_start(
        self, mock_rest_bot: rest_bot_impl.RESTBot, mock_interaction_server: mock.Mock, mock_rest_client: mock.Mock
    ):
        mock_socket = object()
        mock_ssl_context = object()
        mock_callback_1 = mock.AsyncMock()
        mock_callback_2 = mock.AsyncMock()
        mock_rest_bot.add_startup_callback(mock_callback_1)
        mock_rest_bot.add_startup_callback(mock_callback_2)
        mock_rest_bot._is_closing = True

        with mock.patch.object(ux, "check_for_updates"):
            await mock_rest_bot.start(
                backlog=34123,
                check_for_updates=False,
                host="hostostosot",
                port=123123123,
                path="patpatpapt",
                reuse_address=True,
                reuse_port=False,
                socket=mock_socket,
                shutdown_timeout=4312312.3132132,
                ssl_context=mock_ssl_context,
            )

            ux.check_for_updates.assert_not_called()

        mock_interaction_server.start.assert_awaited_once_with(
            backlog=34123,
            host="hostostosot",
            port=123123123,
            path="patpatpapt",
            reuse_address=True,
            reuse_port=False,
            socket=mock_socket,
            shutdown_timeout=4312312.3132132,
            ssl_context=mock_ssl_context,
        )
        mock_rest_client.start.assert_called_once_with()
        mock_rest_client.close.assert_not_called()
        assert mock_rest_bot._is_closing is False
        mock_callback_1.assert_awaited_once_with(mock_rest_bot)
        mock_callback_2.assert_awaited_once_with(mock_rest_bot)

    @pytest.mark.asyncio
    async def test_start_when_startup_callback_raises(
        self, mock_rest_bot: rest_bot_impl.RESTBot, mock_interaction_server: mock.Mock, mock_rest_client: mock.Mock
    ):
        mock_socket = object()
        mock_ssl_context = object()
        mock_rest_bot._is_closing = True
        mock_error = TypeError("Not a real catgirl")
        mock_callback_1 = mock.AsyncMock(side_effect=mock_error)
        mock_callback_2 = mock.AsyncMock()
        mock_rest_bot.add_startup_callback(mock_callback_1)
        mock_rest_bot.add_startup_callback(mock_callback_2)

        with mock.patch.object(ux, "check_for_updates"):
            with pytest.raises(TypeError) as exc_info:
                await mock_rest_bot.start(
                    backlog=34123,
                    check_for_updates=False,
                    host="hostostosot",
                    port=123123123,
                    path="patpatpapt",
                    reuse_address=True,
                    reuse_port=False,
                    socket=mock_socket,
                    shutdown_timeout=4312312.3132132,
                    ssl_context=mock_ssl_context,
                )

            assert exc_info.value is mock_error
            ux.check_for_updates.assert_not_called()

        mock_interaction_server.start.assert_not_called()
        mock_rest_client.start.assert_called_once_with()
        mock_rest_client.close.assert_awaited_once_with()
        assert mock_rest_bot._is_closing is False
        mock_callback_1.assert_awaited_once_with(mock_rest_bot)
        mock_callback_2.assert_not_called()

    @pytest.mark.asyncio
    async def test_start_checks_for_update(self, mock_rest_bot, mock_http_settings, mock_proxy_settings):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(asyncio, "create_task"))
        stack.enter_context(mock.patch.object(ux, "check_for_updates", new=mock.Mock()))

        with stack:
            await mock_rest_bot.start(
                backlog=34123,
                check_for_updates=True,
                host="hostostosot",
                port=123123123,
                path="patpatpapt",
                reuse_address=True,
                reuse_port=False,
                socket=object(),
                shutdown_timeout=4312312.3132132,
                ssl_context=object(),
            )

            asyncio.create_task.assert_called_once_with(
                ux.check_for_updates.return_value, name="check for package updates"
            )
            ux.check_for_updates.assert_called_once_with(mock_http_settings, mock_proxy_settings)

    @pytest.mark.asyncio
    async def test_start_when_is_alive(self, mock_rest_bot):
        mock_rest_bot._close_event = object()

        with mock.patch.object(ux, "check_for_updates", new=mock.Mock()) as check_for_updates:
            with pytest.raises(errors.ComponentStateConflictError):
                await mock_rest_bot.start()

            check_for_updates.assert_not_called()

    def test_get_listener(self, mock_rest_bot, mock_interaction_server):
        mock_type = object()

        result = mock_rest_bot.get_listener(mock_type)

        assert result is mock_interaction_server.get_listener.return_value
        mock_interaction_server.get_listener.assert_called_once_with(mock_type)

    def test_set_listener(self, mock_rest_bot, mock_interaction_server):
        mock_type = object()
        mock_listener = object()

        mock_rest_bot.set_listener(mock_type, mock_listener, replace=True)

        mock_interaction_server.set_listener.assert_called_once_with(mock_type, mock_listener, replace=True)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import ssl

import pytest

from hikari.impl import config as config_


class TestSSLFactory:
    def test_when_value_is_True(self):
        returned = config_._ssl_factory(True)

        assert returned.check_hostname is True
        assert returned.verify_mode is ssl.CERT_REQUIRED

    def test_when_value_is_False(self):
        returned = config_._ssl_factory(False)

        assert returned.check_hostname is False
        assert returned.verify_mode is ssl.CERT_NONE

    def test_when_value_is_non_bool(self):
        value = object()
        assert config_._ssl_factory(value) is value


class TestBasicAuthHeader:
    @pytest.fixture
    def config(self):
        return config_.BasicAuthHeader(username="davfsa", password="securepassword123")

    def test_header_property(self, config):
        assert config.header == f"{config_._BASICAUTH_TOKEN_PREFIX} ZGF2ZnNhOnNlY3VyZXBhc3N3b3JkMTIz"

    def test_str(self, config):
        assert str(config) == f"{config_._BASICAUTH_TOKEN_PREFIX} ZGF2ZnNhOnNlY3VyZXBhc3N3b3JkMTIz"


class TestHTTPTimeoutSettings:
    @pytest.mark.parametrize("arg", ["acquire_and_connect", "request_socket_connect", "request_socket_read", "total"])
    def test_max_redirects_validator_when_not_None_nor_int_nor_float(self, arg):
        with pytest.raises(ValueError, match=rf"HTTPTimeoutSettings.{arg} must be None, or a POSITIVE float/int"):
            config_.HTTPTimeoutSettings(**{arg: object()})

    @pytest.mark.parametrize("arg", ["acquire_and_connect", "request_socket_connect", "request_socket_read", "total"])
    def test_max_redirects_validator_when_negative_int(self, arg):
        with pytest.raises(ValueError, match=rf"HTTPTimeoutSettings.{arg} must be None, or a POSITIVE float/int"):
            config_.HTTPTimeoutSettings(**{arg: -1})

    @pytest.mark.parametrize("arg", ["acquire_and_connect", "request_socket_connect", "request_socket_read", "total"])
    def test_max_redirects_validator_when_negative_float(self, arg):
        with pytest.raises(ValueError, match=rf"HTTPTimeoutSettings.{arg} must be None, or a POSITIVE float/int"):
            config_.HTTPTimeoutSettings(**{arg: -1.1})

    @pytest.mark.parametrize("arg", ["acquire_and_connect", "request_socket_connect", "request_socket_read", "total"])
    @pytest.mark.parametrize("value", [1, 1.1, None])
    def test_max_redirects_validator(self, arg, value):
        config_.HTTPTimeoutSettings(**{arg: value})


class TestHTTPSettings:
    def test_max_redirects_validator_when_not_None_nor_int(self):
        with pytest.raises(ValueError, match=r"http_settings.max_redirects must be None or a POSITIVE integer"):
            config_.HTTPSettings(max_redirects=object())

    def test_max_redirects_validator_when_negative(self):
        with pytest.raises(ValueError, match=r"http_settings.max_redirects must be None or a POSITIVE integer"):
            config_.HTTPSettings(max_redirects=-1)

    @pytest.mark.parametrize("value", [1, None])
    def test_max_redirects_validator(self, value):
        config_.HTTPSettings(max_redirects=value)

    def test_ssl(self):
        mock_ssl = ssl.create_default_context()
        config = config_.HTTPSettings(ssl=mock_ssl)

        assert config.ssl is mock_ssl


class TestProxySettings:
    def test_all_headers_when_headers_and_auth_are_None(self):
        config = config_.ProxySettings(headers=None, auth=None)
        assert config.all_headers is None

    def test_all_headers_when_headers_is_None_and_auth_is_not_None(self):
        config = config_.ProxySettings(headers=None, auth="some auth")
        assert config.all_headers == {config_._PROXY_AUTHENTICATION_HEADER: "some auth"}

    def test_all_headers_when_headers_is_not_None_and_auth_is_None(self):
        config = config_.ProxySettings(headers={"header1": "header1 info"}, auth=None)
        assert config.all_headers == {"header1": "header1 info"}

    def test_all_headers_when_headers_and_auth_are_not_None(self):
        config = config_.ProxySettings(headers={"header1": "header1 info"}, auth="some auth")
        assert config.all_headers == {"header1": "header1 info", config_._PROXY_AUTHENTICATION_HEADER: "some auth"}



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import cProfile
import enum as py_enum
import timeit

from hikari.internal import enums as hikari_enum


class BasicPyEnum(str, py_enum.Enum):
    __slots__ = ()

    a = "0"
    b = "1"
    c = "2"
    d = "3"
    e = "4"
    f = "5"
    g = "6"
    h = "7"
    i = "8"
    j = "9"
    k = "10"
    m = "12"
    n = "13"
    o = "14"
    p = "15"
    q = "16"
    r = "17"
    s = "18"
    t = "19"
    u = "20"
    v = "21"
    w = "22"
    x = "23"
    y = "24"
    z = "25"


class BasicHikariEnum(str, hikari_enum.Enum):
    __slots__ = ()

    a = "0"
    b = "1"
    c = "2"
    d = "3"
    e = "4"
    f = "5"
    g = "6"
    h = "7"
    i = "8"
    j = "9"
    k = "10"
    m = "12"
    n = "13"
    o = "14"
    p = "15"
    q = "16"
    r = "17"
    s = "18"
    t = "19"
    u = "20"
    v = "21"
    w = "22"
    x = "23"
    y = "24"
    z = "25"


# Dummy work to churn the CPU up.
for i in range(100_000):
    assert sum(i for i in range(10)) > 0

py_enum_call_time = timeit.timeit("BasicPyEnum('25')", number=1_000_000, globals=globals())
hikari_enum_call_time = timeit.timeit("BasicHikariEnum('25')", number=1_000_000, globals=globals())
py_enum_delegate_to_map_time = timeit.timeit(
    "BasicPyEnum._value2member_map_['25']", number=1_000_000, globals=globals()
)
hikari_enum_delegate_to_map_time = timeit.timeit(
    "BasicHikariEnum._value_to_member_map_['25']", number=1_000_000, globals=globals()
)
py_enum_getitem_time = timeit.timeit("BasicPyEnum['z']", number=1_000_000, globals=globals())
hikari_enum_getitem_time = timeit.timeit("BasicHikariEnum['z']", number=1_000_000, globals=globals())

print("BasicPyEnum.__call__('25')", py_enum_call_time, "µs")
print("BasicHikariEnum.__call__('25')", hikari_enum_call_time, "µs")
print("BasicPyEnum._value2member_map_['25']", py_enum_delegate_to_map_time, "µs")
print("BasicHikariEnum._value_to_member_map['25']", hikari_enum_delegate_to_map_time, "µs")
print("BasicPyEnum.__getitem__['z']", py_enum_getitem_time, "µs")
print("BasicHikariEnum.__getitem__['z']", hikari_enum_getitem_time, "µs")

print("BasicPyEnum.__call__ profile")
cProfile.runctx("for i in range(1_000_000): BasicPyEnum('25')", globals=globals(), locals=locals())

print("BasicHikariEnum.__call__ profile")
cProfile.runctx("for i in range(1_000_000): BasicHikariEnum('25')", globals=globals(), locals=locals())

print("BasicPyEnum.__getitem__ profile")
cProfile.runctx("for i in range(1_000_000): BasicPyEnum['z']", globals=globals(), locals=locals())

print("BasicHikariEnum.__getitem__ profile")
cProfile.runctx("for i in range(1_000_000): BasicHikariEnum['z']", globals=globals(), locals=locals())
